#include "Rmatlab.h"

SEXP rmxArrayTag;

static inline Rboolean isRmxArray(SEXP ex)
{
	return isNull(ex) || TYPEOF(ex) == EXTPTRSXP && R_ExternalPtrTag(ex) == rmxArrayTag;
}

mxArray *rmxArray(SEXP ex)
{
	return isRmxArray(ex) && !isNull(ex) ? R_ExternalPtrAddr(ex) : NULL;
}

static void rmxDestroyArray(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	SEXP prot = R_ExternalPtrProtected(ex);
	R_ClearExternalPtr(ex);
	if (!mx || isNull(prot))
		return;
	mxDestroyArray(mx);
}

SEXP rmxArrayNew(mxArray *mx, bool allocated)
{
	if (mx == NULL)
		return R_NilValue;
	SEXP ex = R_MakeExternalPtr(mx, rmxArrayTag, allocated ? R_UnboundValue : R_NilValue);
	R_RegisterCFinalizer(ex, rmxDestroyArray);
	return ex;
}

SEXP RmxIsArray(SEXP ex)
{
	return ScalarLogical(isRmxArray(ex));
}

SEXP RmxDuplicateArray(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	if (!mx) return R_NilValue;
	return rmxArrayNew(mxDuplicateArray(mx), true);
}

static inline mxArray *rmxTakeArray(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	if (!mx) return NULL;
	if (0 && NAMED(ex))
		mx = mxDuplicateArray(mx);
	else
		R_SetExternalPtrProtected(ex, R_NilValue);
	return mx;
}

static const char *mxClassNames[] = 
	{ [mxUNKNOWN_CLASS] 	= "unknown"
	, [mxCELL_CLASS] 	= "cell"
	, [mxSTRUCT_CLASS]	= "struct"
	, [mxLOGICAL_CLASS]	= "logical"
	, [mxCHAR_CLASS]	= "char"
	, [mxVOID_CLASS]	= "void"
	, [mxDOUBLE_CLASS]	= "double"
	, [mxSINGLE_CLASS]	= "single"
	, [mxINT8_CLASS]	= "int8"
	, [mxUINT8_CLASS]	= "uint8"
	, [mxINT16_CLASS]	= "int16"
	, [mxUINT16_CLASS]	= "uint16"
	, [mxINT32_CLASS]	= "int32"
	, [mxUINT32_CLASS]	= "uint32"
	, [mxINT64_CLASS]	= "int64"
	, [mxUINT64_CLASS]	= "uint64"
	, [mxFUNCTION_CLASS]	= "function"
	, [mxOPAQUE_CLASS]	= "opaque"
	, [mxOBJECT_CLASS]	= "object"
	};
static const unsigned mxClassCount = sizeof(mxClassNames)/sizeof(*mxClassNames);

SEXP RmxClassNames()
{
	SEXP res;
	unsigned i;
	PROTECT(res = allocVector(STRSXP, mxClassCount-1));
	for (i = 1 /* skip unknown */; i < mxClassCount; i ++)
		if (mxClassNames[i])
			SET_STRING_ELT(res, i-1 /* make 0-based */, mkChar(mxClassNames[i]));
	UNPROTECT(1);
	return res;
}

static mxClassID rmxGetClassID(SEXP class)
{
	SEXP name;
	unsigned i;
	if (isInteger(class))
		i = Rf_asInteger(class);
	else
	{
		PROTECT(name = Rf_asChar(class));
		for (i = 0; i < mxClassCount; i ++)
			if (mxClassNames[i] && !strcmp(CHAR(name), mxClassNames[i]))
				break;
		UNPROTECT(1);
	}
	if (i >= mxClassCount)
		i = 0;
	return i;
}

SEXP RmxGetClass(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	const char *class = NULL;
	mxClassID cid = mx ? mxGetClassID(mx) : 0;
	if (cid >= mxClassCount)
		error("unknown mxArray ClassID %d", cid);
	return mkString(mxClassNames[cid]);
}

SEXP RmxGetLength(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	return ScalarInteger(mx ? mxGetNumberOfElements(mx) : 0);
}

SEXP RmxGetDimension(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	return ScalarInteger(mx ? mxGetNumberOfDimensions(mx) : 0);
}

SEXP RmxGetDimensions(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	SEXP res;
	mwSize n, i;
	const mwSize *d;

	if (!mx) return R_NilValue;
	n = mxGetNumberOfDimensions(mx);
	PROTECT(res = allocVector(INTSXP, n));
	d = mxGetDimensions(mx);
	for (i = 0; i < n; i ++)
		INTEGER(res)[i] = d[i];
	UNPROTECT(1);
	return res;
}

static mwSize *rmxGetDims(SEXP dim, mwSize *ndims)
{
	unsigned i, l, n;
	mwSize *d;

	PROTECT(dim = Rf_coerceVector(dim, INTSXP));
	n = l = length(dim);
	if (n < 2)
		n = 2;
	*ndims = n;
	d = malloc(sizeof(*d)*n);
	for (i = 0; i < l; i ++)
		if ((d[i] = INTEGER(dim)[i]) < 0)
		{
			free(d);
			UNPROTECT(1);
			return NULL;
		}
	UNPROTECT(1);
	for (; i < n; i ++)
		d[i] = 1;
	return d;
}

SEXP RmxSetDimensions(SEXP ex, SEXP dim)
{
	mxArray *mx = rmxArray(ex);
	mwSize n, *d;
	int r;

	if (!mx) return R_NilValue;
	d = rmxGetDims(dim, &n);
	if (!d) error_return("invalid mxArray dimensions");
	r = mxSetDimensions(mx, d, n);
	free(d);
	return ScalarLogical(!r);
}

static long rmxCalcSingleSubscript(const mxArray *mx, SEXP sub)
{
	mwSize n, m;
	unsigned i;
	PROTECT(sub = Rf_coerceVector(sub, INTSXP));
	
	n = length(sub);
	m = mxGetNumberOfDimensions(mx);
	if (n == 0)
	{
		UNPROTECT(1);
		return 0;
	}
	if (n == 1 || n == 2 && INTEGER(sub)[1] == 1)
	{
		i = INTEGER(sub)[0]-1;
		UNPROTECT(1);
		return i;
	}
	if (n > m)
	{
		UNPROTECT(1);
		return -1;
	}
	mwIndex s[n];
	for (i = 0; i < n; i ++)
	{
		int j = INTEGER(sub)[i]-1;
		if (j < 0)
		{
			UNPROTECT(1);
			return -1;
		}
		s[i] = j;
	}
	UNPROTECT(1);
	return mxCalcSingleSubscript(mx, n, s);
}

static inline SEXP rmxIs_fun(bool fun(const mxArray *), SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	return mx ? ScalarLogical(fun(mx)) : R_NilValue;
}

SEXP RmxIsGlobal(SEXP ex)
{
	return rmxIs_fun(mxIsFromGlobalWS, ex);
}

SEXP RmxSetGlobal(SEXP ex, SEXP global)
{
	mxArray *mx = rmxArray(ex);
	if (!mx)
		return R_NilValue;
	bool glob = asLogical(global);
	mxSetFromGlobalWS(mx, glob);
	return ScalarLogical(glob);
}

#define RmxIs_FUN(WHAT) \
SEXP RmxIs##WHAT(SEXP ex) \
	{ return rmxIs_fun(mxIs##WHAT, ex); }

RmxIs_FUN(Complex)
RmxIs_FUN(Sparse)
static RmxIs_FUN(Numeric)
static RmxIs_FUN(Cell)

SEXP RmxGetElement(SEXP ex, SEXP off)
{
	mxArray *mx = rmxArray(ex);
	SEXP res = 0;
	bool cplx;

	if (!mx) return R_NilValue;
	if (mxIsSparse(mx))
		error_return("unsupported sparse mxArray");
	cplx = mxIsComplex(mx);
	long i = rmxCalcSingleSubscript(mx, off);
	if (i < 0)
		error_return("invalid mxArray subscript");
	if (i >= mxGetNumberOfElements(mx))
		error_return("mxArray subscript out of bounds");
	#define GET(SXP, RF, MXT, MXF) \
		PROTECT(res = allocVector(SXP, 1)); \
		RF(res)[0] = ((MXT *)mxGet##MXF(mx))[i]
	#define GETCPLX(MXT) \
		GET(cplx ? CPLXSXP : REALSXP, REAL, MXT, Pr); \
		if (cplx) COMPLEX(res)->i = ((MXT *)mxGetPi(mx))[i]; \
		cplx = false /* handled */
	switch (mxGetClassID(mx))
	{
		case mxDOUBLE_CLASS:	GETCPLX(double); break;
		case mxSINGLE_CLASS:	GETCPLX(float); break;
		case mxINT8_CLASS:  	GET(INTSXP, INTEGER, int8_T, Data); break;
		case mxUINT8_CLASS: 	GET(INTSXP, INTEGER, uint8_T, Data); break;
		case mxINT16_CLASS: 	GET(INTSXP, INTEGER, int16_T, Data); break;
		case mxUINT16_CLASS:	GET(INTSXP, INTEGER, uint16_T, Data); break;
		case mxINT32_CLASS: 	GET(INTSXP, INTEGER, int32_T, Data); break;
		case mxLOGICAL_CLASS:	GET(LGLSXP, LOGICAL, mxLogical, Data); break;

		case mxCELL_CLASS:
			return rmxArrayNew(mxGetCell(mx, i), false);

		case mxCHAR_CLASS:
			PROTECT(res = allocVector(STRSXP, 1));
			char s = mxGetChars(mx)[i]; /* XXX from char-16 */
			SET_STRING_ELT(res, 0, mkCharLen(&s, 1));
			break;
	}
	#undef GET
	#undef GETCPLX
	if (!res) error_return("unsupported mxArray subscript");
	UNPROTECT(1);
	if (cplx) error_return("unsupported complex mxArray");
	return res;
}

SEXP RmxGetRaw(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	SEXP res;

	if (!mx) return R_NilValue;
	size_t s = mxGetElementSize(mx) * mxGetNumberOfElements(mx);
	PROTECT(res = allocVector(RAWSXP, s));
	memcpy(RAW(res), mxGetData(mx), s);
	UNPROTECT(1);
	return res;
}

SEXP RmxGetVector(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	SEXP res;

	if (!mx) return R_NilValue;
	if (mxIsSparse(mx)) return R_NA;
	if (mxIsComplex(mx)) return R_NA;
	mwSize n = mxGetNumberOfElements(mx);
	if (mxIsDouble(mx))
	{
		PROTECT(res = allocVector(REALSXP, n));
		memcpy(REAL(res), mxGetPr(mx), n*sizeof(double));
	}
	else if (mxIsInt32(mx) && sizeof(int32_T) == sizeof(int))
	{
		PROTECT(res = allocVector(INTSXP, n));
		memcpy(INTEGER(res), mxGetPr(mx), n*sizeof(int));
	}
	else if (mxIsLogical(mx) && sizeof(mxLogical) == sizeof(int))
	{
		PROTECT(res = allocVector(LGLSXP, n));
		memcpy(LOGICAL(res), mxGetLogicals(mx), n*sizeof(int));
	}
	else return R_NA;
	UNPROTECT(1);
	return res;
}

SEXP RmxSetElement(SEXP ex, SEXP off, SEXP val)
{
	mxArray *mx = rmxArray(ex);
	bool cplx;

	if (!mx) return R_NilValue;
	if (mxIsSparse(mx))
		error_return("unsupported sparse mxArray");
	cplx = mxIsComplex(mx);
	long i = rmxCalcSingleSubscript(mx, off);
	if (i < 0)
		error_return("invalid mxArray subscript");
	if (i >= mxGetNumberOfElements(mx))
		error_return("mxArray subscript out of bounds");
	#define SET(RT, MXT, MXF) \
		if (cplx) error_return("unsupported complex mxArray"); \
		((MXT *)mxGet##MXF(mx))[i] = Rf_as##RT(val)
	#define SETCPLX(MXT) \
		if (cplx) { \
			Rcomplex v = Rf_asComplex(val); \
			((MXT *)mxGetPr(mx))[i] = v.r; \
			((MXT *)mxGetPi(mx))[i] = v.i; \
			cplx = false; \
		} else { \
			SET(Real, MXT, Pr); \
		}
	switch (mxGetClassID(mx))
	{
		case mxDOUBLE_CLASS:	SETCPLX(double); break;
		case mxSINGLE_CLASS:	SETCPLX(float); break;
		case mxINT8_CLASS:  	SET(Integer, int8_T, Data); break;
		case mxUINT8_CLASS: 	SET(Integer, uint8_T, Data); break;
		case mxINT16_CLASS: 	SET(Integer, int16_T, Data); break;
		case mxUINT16_CLASS:	SET(Integer, uint16_T, Data); break;
		case mxINT32_CLASS: 	SET(Integer, int32_T, Data); break;
		case mxLOGICAL_CLASS:	SET(Logical, mxLogical, Data); break;

		case mxCELL_CLASS:
		{
			mxArray *v, *o;
			if (!isRmxArray(val))
				error_return("invalid mxCellArray assignment");
			v = rmxArray(val);
			o = mxGetCell(mx, i);
			if (o == v)
				break;
			v = rmxTakeArray(val);
			mxSetCell(mx, i, v);
			if (o)
				mxDestroyArray(o);
			break;
		}

		case mxCHAR_CLASS:
			mxGetChars(mx)[i] = CHAR(Rf_asChar(val))[0];
			break;

		default:
			error_return("unsupported mxArray subscript");
	}
	#undef SET
	#undef SETCPLX
	return val;
}

SEXP RmxSetVector(SEXP ex, SEXP vx)
{
	mxArray *mx = rmxArray(ex);

	mwSize n = mx ? mxGetNumberOfElements(mx) : 0;
	if (n != length(vx))
		error_return("incorrect number of replacement elements for mxArray");
	if (!n) return vx;
	if (mxIsSparse(mx)) return R_NA;
	if (mxIsComplex(mx)) return R_NA;
	if (mxIsDouble(mx))
	{
		vx = PROTECT(Rf_coerceVector(vx, REALSXP));
		memcpy(mxGetPr(mx), REAL(vx), n*sizeof(double));
	}
	else if (mxIsInt32(mx) && sizeof(int32_T) == sizeof(int))
	{
		vx = PROTECT(Rf_coerceVector(vx, INTSXP));
		memcpy(mxGetPr(mx), INTEGER(vx), n*sizeof(int));
	}
	else if (mxIsLogical(mx) && sizeof(mxLogical) == sizeof(int))
	{
		vx = PROTECT(Rf_coerceVector(vx, LGLSXP));
		memcpy(mxGetLogicals(mx), LOGICAL(vx), n*sizeof(int));
	}
	else return R_NA;
	UNPROTECT(1);
	return vx;
}

#define rmxCreate_(TYPE, ARGS...) ({ \
	mwSize _n, *_d; \
	mxArray *mx; \
	_d = rmxGetDims(dim, &_n); \
	if (!_d) error_return("invalid mxArray dimensions"); \
	mx = mxCreate##TYPE##Array(_n, _d, ##ARGS); \
	free(_d); \
	rmxArrayNew(mx, true); \
})


SEXP RmxCreateCellArray(SEXP dim)
{
	return rmxCreate_(Cell);
}

SEXP RmxCreateNumericArray(SEXP dim, SEXP class, SEXP complex)
{
	mxClassID c;
	bool cplx;

	c = rmxGetClassID(class);
	if (!c)
		error_return("invalid mxNumericArray class");
	cplx = Rf_asLogical(complex);
	return rmxCreate_(Numeric, c, cplx);
}

SEXP RmxCreateLogicalArray(SEXP dim)
{
	return rmxCreate_(Logical);
}

SEXP RmxCreateCharArray(SEXP dim)
{
	return rmxCreate_(Char);
}

SEXP RmxCreateStructArray(SEXP dim, SEXP fields)
{
	if (TYPEOF(fields) != STRSXP)
		error_return("invalid mxStructArray fields");
	unsigned n = length(fields), i;
	const char *f[n];
	for (i = 0; i < n; i ++)
		f[i] = CHAR(STRING_ELT(fields,i));
	return rmxCreate_(Struct, n, f);
}

SEXP RmxGetString(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	char *s;
	SEXP res;

	if (!mx) return R_NilValue;
	if (!mxIsChar(mx))
		error_return("RmxGetString: not a character array");
	s = mxArrayToString(mx);
	PROTECT(res = mkString(s));
	mxFree(s);
	UNPROTECT(1);
	return res;
}

SEXP RmxCreateStrings(SEXP sx)
{
	mxArray *mx;
	unsigned n;

	if (!isString(sx))
		error_return("RmxCreateString: non-character");
	n = length(sx);
	if (n == 1)
	{
		mx = mxCreateString(CHAR(STRING_ELT(sx,0)));
	}
	else
	{
		const char *s[n];
		unsigned i;
		for (i = 0; i < n; i ++)
			s[i] = CHAR(STRING_ELT(sx,i));
		mx = mxCreateCharMatrixFromStrings(n, s);
	}
	return rmxArrayNew(mx, true);
}

SEXP RmxGetFields(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	if (!mx) return R_NilValue;
	if (!mxIsStruct(mx))
		error_return("RmxGetFields: not a struct array");
	return ScalarInteger(mxGetNumberOfFields(mx));
}

SEXP RmxGetFieldNames(SEXP ex)
{
	mxArray *mx = rmxArray(ex);
	SEXP res;
	int n, i;

	if (!mx) return R_NilValue;
	if (!mxIsStruct(mx))
		error_return("RmxGetFields: not a struct array");
	n = mxGetNumberOfFields(mx);
	PROTECT(res = allocVector(STRSXP, n));
	for (i = 0; i < n; i ++)
		SET_STRING_ELT(res, i, mkChar(mxGetFieldNameByNumber(mx, i)));
	UNPROTECT(1);
	return res;
}

static int rmxGetFieldNumber(const mxArray *mx, SEXP field)
{
	if (isString(field))
		return mxGetFieldNumber(mx, CHAR(Rf_asChar(field)));
	else
		return Rf_asInteger(field)-1;
}

SEXP RmxGetField(SEXP ex, SEXP off, SEXP field)
{
	mxArray *mx = rmxArray(ex);
	SEXP res = 0;

	if (!mx) return R_NilValue;
	if (!mxIsStruct(mx))
		error_return("RmxGetField: not a struct array");
	long i = rmxCalcSingleSubscript(mx, off);
	if (i < 0)
		error_return("invalid mxArray subscript");
	if (i >= mxGetNumberOfElements(mx))
		error_return("mxArray subscript out of bounds");
	int f = rmxGetFieldNumber(mx, field);
	if (f < 0 && f >= mxGetNumberOfFields(mx))
		error_return("mxStructArray field subscript out of bounds");
	return rmxArrayNew(mxGetFieldByNumber(mx, i, f), false);
}

SEXP RmxSetField(SEXP ex, SEXP off, SEXP field, SEXP val)
{
	mxArray *mx = rmxArray(ex);

	if (!mx) return R_NilValue;
	if (!mxIsStruct(mx))
		error_return("RmxSetField: not a struct array");
	long i = rmxCalcSingleSubscript(mx, off);
	if (i < 0)
		error_return("invalid mxArray subscript");
	if (i >= mxGetNumberOfElements(mx))
		error_return("mxArray subscript out of bounds");
	int f = rmxGetFieldNumber(mx, field);
	if (f < 0 && f >= mxGetNumberOfFields(mx))
		error_return("mxStructArray field subscript out of bounds");
	mxArray *v, *o;
	if (!isRmxArray(val))
		error_return("invalid mxStructArray assignment");
	v = rmxArray(val);
	o = mxGetFieldByNumber(mx, i, f);
	if (o == v)
		return val;
	v = rmxTakeArray(val);
	mxSetFieldByNumber(mx, i, f, v);
	if (o)
		mxDestroyArray(o);
	return val;
}

SEXP RmxAddField(SEXP ex, SEXP field)
{
	mxArray *mx = rmxArray(ex);

	if (!mx) return R_NilValue;
	if (!mxIsStruct(mx))
		error_return("RmxAddField: not a struct array");
	int r = mxAddField(mx, CHAR(Rf_asChar(field)));
	if (r >= 0)
		return ScalarInteger(r+1);
	else
		return ScalarLogical(FALSE);
}

SEXP RmxRemoveField(SEXP ex, SEXP field)
{
	mxArray *mx = rmxArray(ex);

	if (!mx) return R_NilValue;
	if (!mxIsStruct(mx))
		error_return("RmxAddField: not a struct array");
	int f = rmxGetFieldNumber(mx, field);
	if (f < 0 && f >= mxGetNumberOfFields(mx))
		error_return("mxStructArray field doesn't exist");
	mxRemoveField(mx, f);
	return R_NilValue;
}

SEXP RmxTestArray()
{
	mwSize d[] = {2,2};
	const char *f[3] = {"a","bc","ab"};
	mxArray *mx = mxCreateCellArray(sizeof(d)/sizeof(*d), d);
	mxSetCell(mx, 0, mxCreateDoubleScalar(3.14)); 
	mxSetCell(mx, 1, mxCreateString("foobar"));
	mxSetCell(mx, 2, mxCreateStructMatrix(2,1,3,f));
	return rmxArrayNew(mx, true);
}
